#  Copyright (c) 2023 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import yaml
from multipledispatch import dispatch
from typing import Any

from galaxy.app.ioc.definition import ComponentDefinition,          \
                                      ConfigurationDefinition,      \
                                      ReferenceDefinition,          \
                                      InnerComponentDefinition,     \
                                      ValueDefinition,              \
                                      DictDefinition,               \
                                      ListDefinition,               \
                                      TupleDefinition,              \
                                      SetDefinition,                \
                                      FrozensetDefinition
from galaxy.app.ioc.ioc import ApplicationContext
from galaxy.app.app import Application
from galaxy.kernel.kernel import Kernel
from galaxy.service.service import ServiceManager,                  \
                                   ServiceAsyncManager,             \
                                   Manager,                         \
                                   AsyncManager,                    \
                                   Service,                         \
                                   AsyncService
from galaxy.net.net import NetworkService,                          \
                           NetworkAsyncService,                     \
                           Client,                                  \
                           AsyncClient,                             \
                           Server,                                  \
                           AsyncServer
from galaxy.utils.pattern import Visitor
from galaxy.utils.type import TypeUtility
from galaxy.proc.proc import ProcessManager,                        \
                             ProcessAsyncManager,                   \
                             Process,                               \
                             AsyncProcess


class PrintVisitor(Visitor):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(PrintVisitor, self).__init__()

    @dispatch(ComponentDefinition)
    def visit(self, def_: ComponentDefinition) -> None:
        self._print_comp_def(def_)

    def _print_comp_def(self, def_: ComponentDefinition, level: int | None = 0) -> None:
        print("{}component = {}".format("\t" * level, def_.id))
        print("{}scope = {}".format("\t" * (level + 1), def_.scope))
        print("{}properties :".format("\t" * (level + 1)))
        for prop in def_.properties:
            prop.accept(self)
        print("")

    @dispatch(ConfigurationDefinition)
    def visit(self, def_: ComponentDefinition) -> None:
        print("component = {}".format(def_.id))
        print("properties :")
        for prop in def_.properties:
            prop.accept(self)
        print("")

    @dispatch(ReferenceDefinition)
    def visit(self, def_: ReferenceDefinition) -> None:
        print("ReferenceDefinition - {} : {}".format(def_.name, def_.ref_comp_id))

    @dispatch(InnerComponentDefinition)
    def visit(self, def_: InnerComponentDefinition) -> None:
        print(str(def_))

    @dispatch(ValueDefinition)
    def visit(self, def_: ValueDefinition) -> None:
        print("ValueDefinition - {} : {}".format(def_.name, def_.value))

    @dispatch(DictDefinition)
    def visit(self, def_: DictDefinition) -> None:
        print(str(def_))

    @dispatch(ListDefinition)
    def visit(self, def_: ListDefinition) -> None:
        print(str(def_))

    @dispatch(TupleDefinition)
    def visit(self, def_: TupleDefinition) -> None:
        print(str(def_))

    @dispatch(SetDefinition)
    def visit(self, def_: SetDefinition) -> None:
        print(str(def_))

    @dispatch(FrozensetDefinition)
    def visit(self, def_: FrozensetDefinition) -> None:
        print(str(def_))


class Quoted(str):
    pass


class YamlStatusVisitor(Visitor):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(YamlStatusVisitor, self).__init__()
        self.res: dict[str, Any] | None = None
        self._init_yaml()

    def _init_yaml(self):
        def quoted_presenter(dumper, data):
            return dumper.represent_scalar("tag:yaml.org,2002:str", data, style='"')

        yaml.add_representer(Quoted, quoted_presenter)
        yaml.representer.SafeRepresenter.add_representer(Quoted, quoted_presenter)

    @dispatch(ApplicationContext)
    def visit(self, obj: ApplicationContext) -> None:
        apps = obj.get_components_by_type(Application)
        res = {}
        if len(apps) > 0:
            self.visit(list(apps.values())[0], depth=True)
            res["status"] = self.res
        self.res = res

    @dispatch(Application, depth=bool)
    def visit(self, obj: Application, depth: bool = False) -> None:
        res = {
               "id": str(obj.id),
               "name": Quoted(obj.name),
               "type": TypeUtility.classname(obj),
               "status": obj.state
              }
        if depth:
            if hasattr(obj, "kernel") and obj.kernel is not None:
                self.visit(obj.kernel)
                res["kernel"] = self.res
            if hasattr(obj, "service") and obj.service is not None:
                self.visit(obj.service, depth=depth)
                res["service_mgr"] = self.res
            if hasattr(obj, "proc") and obj.proc is not None:
                self.visit(obj.proc, depth=depth)
                res["proc_mgr"] = self.res
        self.res = res

    @dispatch((ServiceManager, ServiceAsyncManager), depth=bool)
    def visit(self, obj: ServiceManager | ServiceAsyncManager, depth: bool = False) -> None:
        res = {
               "id": str(obj.id),
               "name": Quoted(obj.name),
               "type": TypeUtility.classname(obj),
               "status": obj.state
              }
        if depth and hasattr(obj, "managers") and obj.managers is not None and len(obj.managers) > 0:
            res["managers"] = []
            for mgr in obj.managers.values():
                self.visit(mgr, depth=depth)
                res["managers"].append(self.res)
        self.res = res

    @dispatch((ProcessManager, ProcessAsyncManager), depth=bool)
    def visit(self, obj: ProcessManager | ProcessAsyncManager, depth: bool = False) -> None:
        res = {
               "id": str(obj.id),
               "name": Quoted(obj.name),
               "type": TypeUtility.classname(obj),
               "status": obj.state
              }
        if depth and hasattr(obj, "procs") and obj.procs is not None and len(obj.procs) > 0:
            res["procs"] = []
            for proc in obj.procs.values():
                self.visit(proc)
                res["procs"].append(self.res)
        self.res = res

    @dispatch((Manager, AsyncManager), depth=bool)
    def visit(self, obj: Manager | AsyncManager, depth: bool = False) -> None:
        res = {
               "id": str(obj.id),
               "name": Quoted(obj.name),
               "type": TypeUtility.classname(obj),
               "status": obj.state
              }
        if depth and hasattr(obj, "services") and obj.services is not None and len(obj.services) > 0:
            res["services"] = []
            for srv in obj.services.values():
                self.visit(srv, depth=depth)
                res["services"].append(self.res)
        self.res = res

    @dispatch((Service, AsyncService), depth=bool)
    def visit(self, obj: Service | AsyncService, depth: bool = False) -> None:
        res = {
               "id": str(obj.id),
               "name": Quoted(obj.name),
               "type": TypeUtility.classname(obj),
               "status": obj.state
              }
        if depth and (isinstance(obj, NetworkService) or isinstance(obj, NetworkAsyncService)):
            if hasattr(obj, "servers") and obj.servers is not None and len(obj.servers) > 0:
                res["servers"] = []
                for srv in obj.servers.values():
                    self.visit(srv)
                    res["servers"].append(self.res)
            if hasattr(obj, "clients") and obj.clients is not None and len(obj.clients) > 0:
                res["clients"] = []
                for client in obj.clients.values():
                    self.visit(client)
                    res["clients"].append(self.res)
        self.res = res

    @dispatch((Kernel, Server, AsyncServer, Client, AsyncClient, Process, AsyncProcess))
    def visit(self, obj: Kernel | Server | AsyncServer | Client | AsyncClient | Process | AsyncProcess) -> None:
        self.res = {
                    "id": str(obj.id),
                    "name": Quoted(obj.name),
                    "type": TypeUtility.classname(obj),
                    "status": obj.state
                   }


class YamlCompositionVisitor(Visitor):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(YamlCompositionVisitor, self).__init__()
        self.res: dict[str, Any] | list | None = None
        self._init_yaml()

    def _init_yaml(self):
        def quoted_presenter(dumper, data):
            return dumper.represent_scalar("tag:yaml.org,2002:str", data, style='"')
        yaml.add_representer(Quoted, quoted_presenter)
        yaml.representer.SafeRepresenter.add_representer(Quoted, quoted_presenter)

    @dispatch(ApplicationContext)
    def visit(self, context: ApplicationContext) -> None:
        res = []
        for comp_def in context.comp_defs.values():
            self.visit(comp_def)
            res.append(self.res)
        self.res = res

    @dispatch(ComponentDefinition)
    def visit(self, comp_def: ComponentDefinition) -> None:
        props = {}
        for prop in comp_def.properties:
            self.visit(prop)
            props.update(self.res)
        self.res = {
                    "component": str(comp_def.id),
                    "class": comp_def.class_,
                    "scope": comp_def.scope,
                    "lazy-init": comp_def.lazy_init,
                    "properties": props
                   }

    @dispatch(ReferenceDefinition)
    def visit(self, prop_def: ReferenceDefinition) -> None:
        self.res = {prop_def.name: "{{ref: {}}}".format(str(prop_def.ref_comp_id))}

    @dispatch(InnerComponentDefinition)
    def visit(self, prop_def: ReferenceDefinition) -> None:
        self.res = {}

    @dispatch(ValueDefinition)
    def visit(self, prop_def: ValueDefinition) -> None:
        if isinstance(prop_def.value, str):
            self.res = {prop_def.name: Quoted(prop_def.value)}
        else:
            self.res = {prop_def.name: prop_def.value}

    @dispatch(DictDefinition)
    def visit(self, prop_def: DictDefinition) -> None:
        self.res = {}

    @dispatch(ListDefinition)
    def visit(self, prop_def: ListDefinition) -> None:
        self.res = {}

    @dispatch(TupleDefinition)
    def visit(self, prop_def: TupleDefinition) -> None:
        self.res = {}

    @dispatch(SetDefinition)
    def visit(self, prop_def: SetDefinition) -> None:
        self.res = {}

    @dispatch(FrozensetDefinition)
    def visit(self, prop_def: FrozensetDefinition) -> None:
        self.res = {}


class YamlConfigurationVisitor(Visitor):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(YamlConfigurationVisitor, self).__init__()
        self.res: dict[str, Any] | list | None = None

    @dispatch(ApplicationContext)
    def visit(self, context: ApplicationContext) -> None:
        res = []
        for conf_def in context.conf_defs.values():
            self.visit(conf_def)
            res.append(self.res)
        self.res = res

    @dispatch(ConfigurationDefinition)
    def visit(self, conf_def: ConfigurationDefinition) -> None:
        self.visit(conf_def.properties)
        props = self.res
        self.res = {
                    "component": str(conf_def.id),
                    "properties": props
                   }

    @dispatch(DictDefinition)
    def visit(self, prop_def: DictDefinition) -> None:
        res = {}
        self._visit_rec(res, prop_def.value)
        self.res = res

    def _visit_rec(self, res: dict, prop_def: dict):
        for name, val in prop_def.items():
            if isinstance(val, str):
                res[name] = Quoted(val)
            elif isinstance(val, dict):
                tmp = {}
                self._visit_rec(tmp, val)
                res[name] = tmp
            else:
                res[name] = val




#  Copyright (c) 2023 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any,                     \
                   TYPE_CHECKING,           \
                   Optional
from dataclasses import dataclass,          \
                        field

from galaxy.app import constant
from galaxy.utils.type import CompId

if TYPE_CHECKING:
    from galaxy.app.ioc.visitor import Visitor
    from galaxy.app.ioc.factory import ReflectiveObjectFactory
    from galaxy.app.ioc.ioc import ComponentContainer


@dataclass(init=True,
           repr=True,
           eq=True,
           order=False,
           unsafe_hash=False,
           frozen=False)
           #match_args=True,
           #kw_only=False,
           #slots=True,
           #weakref_slot=False
class ComponentDefinition(object):
    """
    classdocs
    """
    id: CompId
    class_: str | None = field(init=False)

    # Object Factory
    factory: Optional["ReflectiveObjectFactory"] = field(init=False)

    # Properties
    properties: list = field(init=False, default_factory=list)

    scope: str = field(init=False, default=constant.SINGLETON)
    lazy_init: bool = field(init=False, default=False)
    abstract: bool = field(init=False, default=False)
    _parent_id: CompId | None = field(init=False)

    # Positional constructors
    pos_constr: list = field(init=False, default_factory=list)

    # Named constructors
    named_constr: dict = field(init=False, default_factory=dict)

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __str__(self) -> str:
        return str(self.id)


@dataclass(init=True,
           repr=True,
           eq=True,
           order=False,
           unsafe_hash=False,
           frozen=False)
           #match_args=True,
           #kw_only=False,
           #slots=True,
           #weakref_slot=False
class ConfigurationDefinition(object):
    """
    classdocs
    """
    id: CompId

    # Properties
    properties: Optional["DictDefinition"] = field(init=False)

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __str__(self) -> str:
        return str(self.id)


@dataclass(init=True,
           repr=True,
           eq=True,
           order=False,
           unsafe_hash=False,
           frozen=False)
           #match_args=True,
           #kw_only=False,
           #slots=True,
           #weakref_slot=False
class ReferenceDefinition(object):
    """
    classdocs
    """
    name: str
    ref_comp_id: CompId

    def prefetch(self, container: "ComponentContainer") -> None:
        self.get_value(container)

    def get_value(self, container: "ComponentContainer") -> object:
        return container.get_component(self.ref_comp_id)

    def set_value(self, comp: object, container: "ComponentContainer") -> None:
        setattr(comp, self.name, container.components[self.ref_comp_id])

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __str__(self) -> str:
        return self.name


@dataclass(init=True,
           repr=True,
           eq=True,
           order=False,
           unsafe_hash=False,
           frozen=False)
           #match_args=True,
           #kw_only=False,
           #slots=True,
           #weakref_slot=False
class InnerComponentDefinition(object):
    """
    classdocs
    """
    name: str
    inner_comp_def: ComponentDefinition

    def prefetch(self, container: "ComponentContainer") -> Any:
        self.get_value(container)

    def get_value(self, container: "ComponentContainer") -> Any:
        return container.get_component(self.inner_comp_def.id)

    def set_value(self, comp: Any, container: "ComponentContainer") -> None:
        setattr(comp, self.name, self.get_value(container))

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __str__(self) -> str:
        return str(self.name)


@dataclass(init=False,
           repr=True,
           eq=True,
           order=False,
           unsafe_hash=False,
           frozen=False)
           #match_args=True,
           #kw_only=False,
           #slots=True,
           #weakref_slot=False
class ValueDefinition(object):
    """
    classdocs
    """
    name: str

    def __init__(self, name: str, value: Any) -> None:
        """
        Constructor
        """
        self.name: str = name
        if value == "True":
            self.value: bool = True
        elif value == "False":
            self.value: bool = False
        else:
            self.value = value

    def scan_value(self, container: "ComponentContainer", value: Any) -> Any:
        if hasattr(value, "get_value"):
            return value.get_value(container)
        elif isinstance(value, tuple):
            new_list = [self.scan_value(container, item) for item in value]
            results = tuple(new_list)
            return results
        elif isinstance(value, list):
            new_list = [self.scan_value(container, item) for item in value]
            return new_list
        elif isinstance(value, set):
            results = set([self.scan_value(container, item) for item in value])
            return results
        elif isinstance(value, frozenset):
            results = frozenset([self.scan_value(container, item) for item in value])
            return results
        else:
            if value == "True":
                return True
            elif value == "False":
                return False
            else:
                return value

    def get_value(self, container: "ComponentContainer") -> Any:
        val = self._replace_refs(self.value, container)
        if val is None:
            return self.value
        else:
            return val

    def set_value(self, comp: Any, container: "ComponentContainer") -> None:
        setattr(comp, self.name, self.value)
        self._replace_refs(comp, container)

    def _replace_refs(self, comp: Any, container: Any):
        """Normal values do nothing for this step. However, sub-classes are defined for
        the various containers, like lists, set, dictionaries, etc., to handle iterating
        through and pre-fetching items."""
        pass

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __str__(self) -> str:
        return self.name


@dataclass(init=False,
           repr=True,
           eq=True,
           order=False,
           unsafe_hash=False,
           frozen=False)
           #match_args=True,
           #kw_only=False,
           #slots=True,
           #weakref_slot=False
class DictDefinition(ValueDefinition):
    """
    classdocs
    """

    def __init__(self, name: str, value: Any) -> None:
        """
        Constructor
        """
        super(DictDefinition, self).__init__(name, value)

    def _replace_refs(self, comp: Any, container: "ComponentContainer"):
        for key in self.value.keys():
            if hasattr(self.value[key], "ref"):
                self.value[key] = container.get_component(self.value[key].ref)
            else:
                self.value[key] = self.scan_value(container, self.value[key])

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __str__(self) -> str:
        return self.name


@dataclass(init=False,
           repr=True,
           eq=True,
           order=False,
           unsafe_hash=False,
           frozen=False)
           #match_args=True,
           #kw_only=False,
           #slots=True,
           #weakref_slot=False
class ListDefinition(ValueDefinition):
    """
    classdocs
    """

    def __init__(self, name: str, value: Any) -> None:
        """
        Constructor
        """
        super(ListDefinition, self).__init__(name, value)

    def _replace_refs(self, comp: Any, container: "ComponentContainer"):
        for i in range(0, len(self.value)):
            if hasattr(self.value[i], "ref"):
                self.value[i] = container.get_component(self.value[i].ref)
            else:
                self.value[i] = self.scan_value(container, self.value[i])

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __str__(self) -> str:
        return self.name


@dataclass(init=False,
           repr=True,
           eq=True,
           order=False,
           unsafe_hash=False,
           frozen=False)
           #match_args=True,
           #kw_only=False,
           #slots=True,
           #weakref_slot=False
class TupleDefinition(ValueDefinition):
    """
    classdocs
    """

    def __init__(self, name: str, value: Any) -> None:
        """
        Constructor
        """
        super(TupleDefinition, self).__init__(name, value)

    def _replace_refs(self, comp: Any, container: "ComponentContainer"):
        new_value = list(self.value)
        for i in range(0, len(new_value)):
            if hasattr(new_value[i], "ref"):
                new_value[i] = container.get_component(new_value[i].ref)
            else:
                new_value[i] = self.scan_value(container, new_value[i])
        try:
            setattr(comp, self.name, tuple(new_value))
        except AttributeError:
            pass
        return tuple(new_value)

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __str__(self) -> str:
        return self.name


@dataclass(init=False,
           repr=True,
           eq=True,
           order=False,
           unsafe_hash=False,
           frozen=False)
           #match_args=True,
           #kw_only=False,
           #slots=True,
           #weakref_slot=False
class SetDefinition(ValueDefinition):
    """
    classdocs
    """

    def __init__(self, name: str, value: Any) -> None:
        """
        Constructor
        """
        super(SetDefinition, self).__init__(name, value)

    def _replace_refs(self, comp: Any, container: "ComponentContainer"):
        new_set = set()
        for item in self.value:
            if hasattr(item, "ref"):
                newly_fetched_value = container.get_component(item.ref)
                new_set.add(newly_fetched_value)
            else:
                newly_scanned_value = self.scan_value(container, item)
                new_set.add(newly_scanned_value)
        try:
            setattr(comp, self.name, new_set)
        except AttributeError:
            pass
        return new_set

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __str__(self) -> str:
        return self.name


@dataclass(init=False,
           repr=True,
           eq=True,
           order=False,
           unsafe_hash=False,
           frozen=False)
           #match_args=True,
           #kw_only=False,
           #slots=True,
           #weakref_slot=False
class FrozensetDefinition(ValueDefinition):
    """
    classdocs
    """

    def __init__(self, name: str, value: Any) -> None:
        """
        Constructor
        """
        super(FrozensetDefinition, self).__init__(name, value)

    def _replace_refs(self, comp: Any, container: "ComponentContainer"):
        new_set = set()
        for item in self.value:
            if hasattr(item, "ref"):
                newly_fetched_value = container.get_component(item.ref)
                new_set.add(newly_fetched_value)
            else:
                newly_scanned_value = self.scan_value(container, item)
                new_set.add(newly_scanned_value)
        new_frozenset = frozenset(new_set)
        try:
            setattr(comp, self.name, new_frozenset)
        except AttributeError:
            pass
        except TypeError:
            pass
        return new_frozenset

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __str__(self) -> str:
        return self.name



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from importlib import import_module
from typing import Any
import sys

from galaxy.app import constant
from galaxy.app.ioc.builder import CompDefinitionBuilder,               \
                                   ConfDefinitionBuilder,               \
                                   RefDefinitionBuilder
from galaxy.app.ioc.definition import ComponentDefinition,              \
                                      ConfigurationDefinition,          \
                                      ReferenceDefinition,              \
                                      ValueDefinition,                  \
                                      InnerComponentDefinition,         \
                                      DictDefinition,                   \
                                      ListDefinition,                   \
                                      TupleDefinition,                  \
                                      SetDefinition,                    \
                                      FrozensetDefinition

yaml_mappings: dict[str, str] = {
                                 "str": "types.StringType",
                                 "unicode": "types.UnicodeType",
                                 "int": "types.IntType",
                                 "long": "types.LongType",
                                 "float": "types.FloatType",
                                 "decimal": "decimal.Decimal",
                                 "bool": "types.BooleanType",
                                 "complex": "types.ComplexType",
                                 "list": "types.ListType",
                                 "tuple": "types.TupleType",
                                 "dict": "types.DictType",
                                }


class DefinitionFactory(object):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        self.comp_defs: list[ComponentDefinition] = []
        self.conf_defs: list[ConfigurationDefinition] = []
        self.abstract_comp_defs: dict = {}

    def create_components(self, def_: dict) -> list[ComponentDefinition]:
        # A dictionary of abstract objects, keyed by their IDs, used in
        # traversing the hierarchies of parents; built upfront here for
        # convenience
        for comp in def_["compo"]:
            if "abstract" in comp:
                self.abstract_comp_defs[comp["component"]] = comp

        for comp in def_["compo"]:
            self.comp_defs.append(self._create_comp_def(comp))
        return self.comp_defs

    def create_configurations(self, def_: dict) -> list[ConfigurationDefinition]:
        for conf in def_["config"]:
            self.conf_defs.append(self._create_conf_def(conf))
        return self.conf_defs

    def _map_custom_class(self, comp: dict, mappings: dict[str, str]) -> None:
        for class_name in mappings:
            if class_name in comp:
                comp["class"] = mappings[class_name]
                comp["constructor-args"] = [comp[class_name]]
                break

    def _create_comp_def(self, comp: dict[str, str], prefix: str = "") -> ComponentDefinition:
        if "class" not in comp and "parent" not in comp:
            self._map_custom_class(comp, yaml_mappings)
        builder = CompDefinitionBuilder().id_(comp["component"])                            \
                                         .class_(comp["class"])                             \
                                         .factory(ReflectiveObjectFactory(comp["class"]))   \
                                         .lazy_init(comp.get("lazy-init", False))           \
                                         .abstract(comp.get("abstract", False))
        if "scope" in comp:
            builder.scope(DefinitionFactory._convert_scope(comp["scope"]))
        if "parent" in comp:
            return self._create_child_comp_def(comp,
                                               comp,
                                               self._get_pos_constr(comp),
                                               self._get_named_constr(comp),
                                               self._get_comp_properties(comp))
        else:
            builder.pos_constr(self._get_pos_constr(comp))              \
                   .named_constr(self._get_named_constr(comp))          \
                   .properties(self._get_comp_properties(comp))
            return builder.build()

    def _create_conf_def(self, conf: dict[str, str], prefix: str = "") -> ConfigurationDefinition:
        return ConfDefinitionBuilder().id_(conf["component"])                       \
                                      .properties(self._get_conf_properties(conf))       \
                                      .build()

    @staticmethod
    def _convert_scope(scope: str) -> str:
        if scope == "prototype":
            return constant.PROTOTYPE
        elif scope == "singleton":
            return constant.SINGLETON
        else:
            raise Exception("Can not handle scope {}".format(scope))

    def _create_child_comp_def(self,
                               leaf: dict[str, str],
                               child_comp: dict[str, str],
                               pos_constr: list,
                               named_constr: dict,
                               properties: list) -> ComponentDefinition:
        parent = self.abstract_comp_defs[child_comp["parent"]]

        # At this point we only build up the lists of parameters but we don't create
        # the object yet because the current parent object may still have its
        # own parent.

        # Positional constructors
        parent_pos_constrs = self._get_pos_constr(parent)

        # Make sure there are as many child positional parameters as there
        # are in the parent's list
        if len(pos_constr) < len(parent_pos_constrs):
            pos_constr.extend([None] * (len(parent_pos_constrs) - len(pos_constr)))

        for idx, parent_pos_constr in enumerate(parent_pos_constrs):
            if not pos_constr[idx]:
                pos_constr[idx] = parent_pos_constr

        # Named constructors
        child_named_constrs = named_constr
        parent_named_constrs = self._get_named_constr(parent)

        for parent_named_constr in parent_named_constrs:
            if parent_named_constr not in child_named_constrs:
                named_constr[parent_named_constr] = parent_named_constrs[parent_named_constr]

        # Properties
        child_properties = [prop.name for prop in properties]
        for parent_prop in self._get_comp_properties(parent):
            if parent_prop.name not in child_properties:
                properties.append(parent_prop)

        if "parent" in parent:
            return self._create_child_comp_def(leaf, parent, pos_constr, named_constr, properties)
        else:
            builder = CompDefinitionBuilder().id_(leaf["component"])                                \
                                             .class_(leaf["class"])                                 \
                                             .factory(ReflectiveObjectFactory(parent["class"]))     \
                                             .lazy_init(leaf.get("lazy-init", False))               \
                                             .abstract(leaf.get("abstract", False))                 \
                                             .pos_constr(pos_constr)                                \
                                             .named_constr(named_constr)                            \
                                             .properties(properties)
            if "scope" in leaf:
                builder.scope(self._convert_scope(leaf["scope"]))
            return builder.build()

    def _get_pos_constr(self, comp: dict) -> list | None:
        if "constructor-args" in comp and isinstance(comp["constructor-args"], list):
            return [self._create_prop_def(comp, constr, comp["object"]) for constr in comp["constructor-args"]]
        return None

    def _get_named_constr(self, comp: dict) -> dict | None:
        if "constructor-args" in comp and isinstance(comp["constructor-args"], dict):
            return dict([(name, self._create_prop_def(comp, constr, comp["object"])) for (name, constr) in comp["constructor-args"].items()])
        return None

    def _get_comp_properties(self, comp: dict) -> list | None:
        if "properties" in comp:
            return [self._create_prop_def(comp, p, name) for (name, p) in comp["properties"].items()]
        return None

    def _get_conf_properties(self, conf: dict) -> DictDefinition | None:
        if "properties" in conf:
            d = {}
            for (name, p) in conf["properties"].items():
                if isinstance(p, dict):
                    d[name] = self._create_prop_def(conf, p, "new_conf.dict['{}']".format(name))
                else:
                    d[name] = self._create_value_def(p, conf["component"], "new_conf.dict['{}']".format(name))
            return DictDefinition("new_conf", d)
        return None

    def _create_ref_def(self, ref_node: Any, name: str) -> ReferenceDefinition:
        if "component" in ref_node:
            return ReferenceDefinition(name, ref_node["component"])
        else:
            builder = RefDefinitionBuilder().name(name).ref_comp_id(ref_node)
            return builder.build()

    def _create_value_def(self, value_node: Any, id_: str, name: str) -> ValueDefinition:
        if isinstance(value_node, dict):
            if "tuple" in value_node:
                return self._create_tuple_def(value_node["tuple"], id_, name)
            elif "list" in value_node:
                return self._create_list_def(value_node["list"], id_, name)
            elif "dict" in value_node:
                return self._create_dict_def(value_node["dict"], id_, name)
            elif "set" in value_node:
                return self._create_set_def(value_node["set"], id_, name)
            elif "frozenset" in value_node:
                return self._create_frozenset_def(value_node["frozenset"], id_, name)
        else:
            return value_node

    def _create_dict_def(self, dict_node: Any, id_: str, name: str) -> DictDefinition:
        d = {}
        for (k, v) in dict_node.items():
            if isinstance(v, dict):
                if "ref" in v:
                    d[k] = self._create_ref_def(v["ref"], "{}.dict['{}']".format(name, k))
                elif "tuple" in v:
                    d[k] = self._create_tuple_def(v["tuple"], id_, "{}.dict['{}']".format(name, k))
                else:
                    d[k] = self._create_dict_def(v, id_, "{}.dict['{}']".format(name, k))
            else:
                d[k] = self._create_value_def(v, id_, "{}.dict['{}']".format(name, k))
        return DictDefinition(name, d)

    def _create_list_def(self, list_node: list, id_: str, name: str) -> ListDefinition:
        list_ = []
        for item in list_node:
            if isinstance(item, dict):
                if "ref" in item:
                    list_.append(self._create_ref_def(item["ref"], "{}.list[{}]".format(name, len(list_))))
                elif "object" in item:
                    list_.append(self._create_inner_comp_def(item, id_, "{}.list[{}]".format(name, len(list_))))
                elif len({"dict", "tuple", "set", "frozenset", "list"} & set(item)) > 0:
                    list_.append(self._create_value_def(item, id_, "{}.list[{}]".format(name, len(list_))))
            else:
                list_.append(item)
        return ListDefinition(name, list_)

    def _create_tuple_def(self, tuple_node: tuple, id_: str, name: str) -> TupleDefinition:
        list_ = []
        for item in tuple_node:
            if isinstance(item, dict):
                if "ref" in item:
                    list_.append(self._create_ref_def(item["ref"], name + ".tuple"))
                elif "object" in item:
                    list_.append(self._create_inner_comp_def(item, id_, "{}.tuple[{}]".format(name, len(list_))))
                elif len({"dict", "tuple", "set", "frozenset", "list"} & set(item)) > 0:
                    list_.append(self._create_value_def(item, id_, "{}.tuple[{}]".format(name, len(list_))))
            else:
                list_.append(item)
        return TupleDefinition(name, tuple(list_))

    def _create_set_def(self, set_node: set, id_: str, name: str) -> SetDefinition:
        s = set()
        for item in set_node:
            if isinstance(item, dict):
                if "ref" in item:
                    s.add(self._create_ref_def(item["ref"], name + ".set"))
                elif "object" in item:
                    s.add(self._create_inner_comp_def(item, id_, "{}.set[{}]".format(name, len(s))))
                elif len({"dict", "tuple", "set", "frozenset", "list"} & set(item)) > 0:
                    s.add(self._create_value_def(item, id_, "{}.set[{}]".format(name, len(s))))
            else:
                s.add(item)
        return SetDefinition(name, s)

    def _create_frozenset_def(self, frozenset_node: set, id_: str, name: str) -> FrozensetDefinition:
        item = self._create_set_def(frozenset_node, id_, name)
        return FrozensetDefinition(name, frozenset(item.value))

    def _create_inner_comp_def(self, comp: dict, id_: str, name: str) -> InnerComponentDefinition:
        inner_comp_def = self._create_comp_def(comp, prefix="{}.{}".format(id_, name))
        self.comp_defs.append(inner_comp_def)
        return InnerComponentDefinition(name, inner_comp_def)

    def _create_prop_def(self, comp: dict, properties: Any, name: str) -> Any:
        # This function translates object properties into useful collections of information for the container.
        if isinstance(properties, dict):
            if "ref" in properties:
                return self._create_ref_def(properties["ref"], name)
            elif "tuple" in properties:
                return self._create_tuple_def(properties["tuple"], comp["component"], name)
            elif "set" in properties:
                return self._create_set_def(properties["set"], comp["component"], name)
            elif "frozenset" in properties:
                return self._create_frozenset_def(properties["frozenset"], comp["component"], name)
            elif "object" in properties:
                return self._create_inner_comp_def(properties, comp["component"], name)
            else:
                return self._create_dict_def(properties, comp["component"], name)
        elif isinstance(properties, list):
            return self._create_list_def(properties, comp["component"], name)
        else:
            return ValueDefinition(name, properties)


class ObjectFactory(object):
    """
    classdocs
    """

    def create(self, constr, named_constr):
        raise NotImplementedError()


class ReflectiveObjectFactory(ObjectFactory):
    """
    classdocs
    """

    def __init__(self, module_and_class: str) -> None:
        """
        Constructor
        """
        self.module_and_class: str = module_and_class

    def create(self, constr: tuple, named_constr: dict) -> object:
        parts = self.module_and_class.split(".")
        module_name = ".".join(parts[:-1])
        class_name = parts[-1]
        if module_name == "":
            return import_module(class_name)(*constr, **named_constr)
        else:
            import_module(module_name)
            cls = getattr(sys.modules[module_name], class_name)
            try:
                o = cls(*constr, **named_constr)
            except Exception as e:
                print(e)
            return o

    def __str__(self) -> str:
        return "ReflectiveObjectFactory({})".format(self.module_and_class)


class PythonObjectFactory(ObjectFactory):
    """
    classdocs
    """

    def __init__(self, method, wrapper) -> None:
        """
        Constructor
        """
        self.method = method
        self.wrapper = wrapper

    def create(self, constr: list, named_constr: dict) -> Any:

        # Setting wrapper's top_func can NOT be done earlier than this method call,
        # because it is tied to a wrapper decorator, which may not have yet been
        # generated.
        self.wrapper.func_globals["top_func"] = self.method.func_name

        # Because @object-based objects use direct code to specify arguments, and NOT
        # external configuration data, this factory doesn't care about the incoming arguments.

        return self.method()

    def __str__(self) -> str:
        return "PythonObjectFactory({})".format(self.method)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import uuid
from abc import ABC,                                                        \
                abstractmethod
from collections import OrderedDict
import atexit
from os import path
from typing import Any,                                                     \
                   TYPE_CHECKING

from galaxy.app import constant
from galaxy.app.ioc.parser import YamlCompositionParser,                    \
                                  CompositionParser,                        \
                                  YamlConfigurationParser,                  \
                                  ConfigurationParser
from galaxy.app.ioc.definition import ComponentDefinition,                  \
                                      ConfigurationDefinition
from galaxy.utils.base import Component
from galaxy.error.app import AppReferenceNotFoundError,                     \
                             AppScopeNotSupportedError,                     \
                             AppAbstractCompCannotBeInstantiatedError
from galaxy.error.io import IOFormatNotSupportedError
from galaxy.utils.type import CompId

if TYPE_CHECKING:
    from galaxy.app.ioc.visitor import Visitor


class IOCManager(Component):
    """
    classdocs
    """

    def __init__(self, app: object) -> None:
        """
        Constructor
        """
        super(IOCManager, self).__init__()
        self.name: str = "IOC Manager"
        self.comp_parsers: dict[str, CompositionParser] = {
                                                           "yml": YamlCompositionParser()
                                                          }
        self.conf_parsers: dict[str, ConfigurationParser] = {
                                                             "yml": YamlConfigurationParser()
                                                            }
        self.context: ApplicationContext = ApplicationContext()
        self.init_comps = {
                           uuid.uuid4(): app,
                           uuid.uuid4(): self
                          }
        self.context.add_components(self.init_comps)

    def load_comp(self, comp_file) -> None:
        extension = path.splitext(comp_file)[1][1:]
        if extension in self.comp_parsers:
            self.context.import_comp_file(comp_file, self.comp_parsers[extension])
        else:
            raise IOFormatNotSupportedError(extension)

    def load_conf(self, conf_file) -> None:
        extension = path.splitext(conf_file)[1][1:]
        if extension in self.conf_parsers:
            self.context.import_conf_file(conf_file, self.conf_parsers[extension])
        else:
            raise IOFormatNotSupportedError(extension)

    def __repr__(self) -> str:
        return "<IOCManager(id='{}')>".format(self.id)


class InitializingObject(ABC):
    """
    classdocs
    """

    @abstractmethod
    def after_properties_set(self) -> None:
        raise NotImplementedError("Should implement after_properties_set()")


class ObjectPostProcessor(ABC):
    """
    classdocs
    """

    @abstractmethod
    def post_process_before_initialization(self, comp: object, id_: CompId) -> object:
        raise NotImplementedError("Should implement post_process_before_initialization()")

    @abstractmethod
    def post_process_after_initialization(self, comp: object, id_: CompId) -> object:
        raise NotImplementedError("Should implement post_process_after_initialization()")


class DisposableObject(ABC):
    """
    classdocs
    """

    @abstractmethod
    def destroy(self) -> None:
        raise NotImplementedError("Should implement destroy()")


class ComponentContainer(object):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        self.comp_defs: dict[CompId, ComponentDefinition] = OrderedDict({})
        self.components: dict[CompId, Component] = OrderedDict({})
        self.conf_defs: dict[CompId, ConfigurationDefinition] = OrderedDict({})

    def get_component(self, id_: CompId, ignore_abstract: bool | None = False) -> object:
        try:
            comp_def = self.comp_defs[id_]
            if comp_def.abstract and not ignore_abstract:
                raise AppAbstractCompCannotBeInstantiatedError(comp_def)
            return self.components[id_]

        except KeyError as e:
            try:
                comp_def = self.comp_defs[id_]
                if comp_def.abstract and not ignore_abstract:
                    raise AppAbstractCompCannotBeInstantiatedError(comp_def)
                comp = self._create_component(comp_def)

                # Evaluate any scopes, and store appropriately.
                if comp_def.scope == constant.SINGLETON:
                    self.components[id_] = comp
                elif comp_def.scope == constant.PROTOTYPE:
                    pass
                else:
                    raise AppScopeNotSupportedError(comp_def)

                return comp
            except KeyError as e:
                raise AppReferenceNotFoundError(id_, internal=e)

    def _get_constructors_pos(self, comp_def: ComponentDefinition) -> tuple:
        if comp_def.pos_constr is None:
            return tuple()
        return tuple([constr.get_value(self) for constr in comp_def.pos_constr if hasattr(constr, "get_value")])

    def _get_constructors_kw(self, comp_def: ComponentDefinition) -> dict:
        if comp_def.named_constr is None:
            return dict()
        return dict([(key, comp_def.named_constr[key].get_value(self)) for key in comp_def.named_constr
                     if hasattr(comp_def.named_constr[key], "get_value")])

    def _create_component(self, comp_def: ComponentDefinition) -> Any:
        if comp_def.pos_constr is not None:
            [constr.prefetch(self) for constr in comp_def.pos_constr if hasattr(constr, "prefetch")]
        if comp_def.named_constr is not None:
            [constr.prefetch(self) for constr in comp_def.named_constr.values() if hasattr(constr, "prefetch")]
        if comp_def.properties is not None:
            [prop.prefetch(self) for prop in comp_def.properties if hasattr(prop, "prefetch")]

        # Res up an instance of the object, with ONLY constructor-based properties set.
        obj = comp_def.factory.create(self._get_constructors_pos(comp_def),
                                      self._get_constructors_kw(comp_def))

        comp: Component = obj
        comp.id = comp_def.id

        # Fill in the other property values.
        if comp_def.properties is not None:
            [prop.set_value(comp, self) for prop in comp_def.properties if hasattr(prop, "set_value")]
        return comp

    def apply_comp_props(self, obj: Any, comp_def: ComponentDefinition) -> None:
        [prop.set_value(obj, self) for prop in comp_def.properties if hasattr(prop, "set_value")]

    def apply_conf_props(self, obj: Any, conf_def: ConfigurationDefinition) -> None:
        conf_def.properties.set_value(obj, self)

    def add_components(self, comps: dict[CompId, object]):
        self.components.update(comps)


class ApplicationContext(ComponentContainer):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ApplicationContext, self).__init__()
        atexit.register(self.shutdown_hook)

    def import_comp_file(self, comp_file: str, parser: CompositionParser) -> None:
        comp_defs = parser.parse(comp_file)
        self._update_preexisting_comp(list(comp_defs.values()))

        for id_, comp_def in comp_defs.items():
            if id_ in self.comp_defs:
                for prop in comp_def.properties:
                    existing_props = [p for p in self.comp_defs[id_].properties if p.name == prop.name]
                    if len(existing_props) == 1:
                        existing_props[0].value = prop.set_value
                    else:
                        self.comp_defs[id_].properties.append(prop)
            else:
                self.comp_defs[id_] = comp_def

            if not comp_def.lazy_init:
                if comp_def.id not in self.components:
                    self.get_component(comp_def.id, ignore_abstract=True)
                else:
                    self.apply_comp_props(self.components[comp_def.id], comp_def)

        comps = [comp for comp in self.components.values() if isinstance(comp, ObjectPostProcessor)]
        post_processors: list[ObjectPostProcessor] = comps

        for id_, comp in self.components.items():
            if not isinstance(comp, ObjectPostProcessor):
                for post_processor in post_processors:
                    self.components[id_] = post_processor.post_process_before_initialization(comp, id_)

        for comp in self.components.values():
            self._apply(comp)

        for comp_name, comp in self.components.items():
            if not isinstance(comp, ObjectPostProcessor):
                for post_processor in post_processors:
                    self.components[comp_name] = post_processor.post_process_after_initialization(comp, comp_name)

    def import_conf_file(self, conf_file: str, parser: ConfigurationParser) -> None:
        self.conf_defs = parser.parse(conf_file)
        for id_, conf_def in self.conf_defs.items():
            if id_ in self.components:
                self.apply_conf_props(self.components[conf_def.id], conf_def)

    def _update_preexisting_comp(self, comp_defs: list[ComponentDefinition]):
        for comp_def in comp_defs:
            found_comp = ()
            for id_, comp in self.components.items():
                comp_class_name = comp_def.class_.split(".")[-1]
                if comp.__class__.__name__ == comp_class_name:
                    found_comp = (id_, comp)
                    break
            if len(found_comp) == 2:
                self.components.pop(found_comp[0])
                found_comp[1].id = comp_def.id
                self.components[comp_def.id] = found_comp[1]

    def _apply(self, obj: Any) -> None:
        if hasattr(obj, "after_properties_set"):
            obj.after_properties_set()
        if hasattr(obj, "post_process_after_initialization"):
            obj.post_process_after_initialization(self)
        if hasattr(obj, "set_app_context"):
            obj.set_app_context(self)

    def get_components_by_type(self, type_: Any, include_type: bool = True) -> dict[CompId, object]:
        result = {}
        for id_, comp in self.components.items():
            if isinstance(comp, type_):
                if not include_type and type(comp) is type_:
                    continue
                result[id_] = comp
        return result

    def shutdown_hook(self) -> None:
        for id_, comp in self.components.items():
            if isinstance(comp, DisposableObject):
                try:
                    if hasattr(comp, "destroy_method"):
                        destroy_method_name = getattr(comp, "destroy_method")
                    else:
                        destroy_method_name = "destroy"

                    destroy_method = getattr(comp, destroy_method_name)

                except Exception:
                    pass

                else:
                    if callable(destroy_method):
                        try:
                            destroy_method()
                        except Exception:
                            pass

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)


class ApplicationContextAware(object):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        self.context: ApplicationContext | None = None

    def set_app_context(self, context: ApplicationContext) -> None:
        self.context = context


class ObjectNameAutoProxyCreator(ApplicationContextAware, ObjectPostProcessor):
    """
    classdocs
    """

    def __init__(self, object_names: list[str] | None = None, interceptor_names: list[str] | None = None) -> None:
        """
        Constructor
        """
        super(ObjectNameAutoProxyCreator, self).__init__()
        self.object_names: list[str] = object_names
        self.interceptor_names: list[str] = interceptor_names



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# State
STATE_NEW = "NEW"
STATE_RUNNING = "RUNNING"
STATE_TIMEOUT = "TIMEOUT"
STATE_STOPPED = "STOPPED"
STATE_SHUTDOWN = "SHUTDOWN"

# IOC Scopes
PROTOTYPE = "prototype"
SINGLETON = "singleton"



#  Copyright (c) 2023 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import uuid
from typing import Any

from galaxy.utils.type import CompId
from galaxy.app.ioc.definition import ComponentDefinition,          \
                                      ConfigurationDefinition,      \
                                      ReferenceDefinition,          \
                                      DictDefinition
from galaxy.utils.pattern import Builder


class CompDefinitionBuilder(Builder):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(CompDefinitionBuilder, self).__init__()
        self._id: CompId | None = None
        self._class: str | None = None
        self._factory: Any | None = None
        self._properties: list = []
        self._scope: str | None = None
        self._lazy_init: bool = False
        self._abstract: bool = False
        self._parent_id: CompId | None = None
        self._pos_constr: list = []
        self._named_constr: dict = {}

    def id_(self, id_: str) -> "CompDefinitionBuilder":
        self._id = uuid.UUID(id_)
        return self

    def class_(self, class_: str) -> "CompDefinitionBuilder":
        self._class = class_
        return self

    def factory(self, factory: "ReflectiveObjectFactory") -> "CompDefinitionBuilder":
        self._factory = factory
        return self

    def properties(self, properties: list) -> "CompDefinitionBuilder":
        self._properties = properties
        return self

    def scope(self, scope: str) -> "CompDefinitionBuilder":
        self._scope = scope
        return self

    def lazy_init(self, lazy_init: bool) -> "CompDefinitionBuilder":
        self._lazy_init = lazy_init
        return self

    def abstract(self, abstract: bool) -> "CompDefinitionBuilder":
        self._abstract = abstract
        return self

    def parent_id(self, parent_id: CompId) -> "CompDefinitionBuilder":
        self._parent_id = parent_id
        return self

    def pos_constr(self, pos_constr: list) -> "CompDefinitionBuilder":
        self._pos_constr = pos_constr
        return self

    def named_constr(self, named_constr: dict) -> "CompDefinitionBuilder":
        self._named_constr = named_constr
        return self

    def build(self) -> ComponentDefinition:
        comp = ComponentDefinition(self._id)
        comp.class_ = self._class
        comp.factory = self._factory
        comp.properties = self._properties
        comp.scope = self._scope
        comp.lazy_init = self._lazy_init
        comp.abstract = self._abstract
        comp.parent_id = self._parent_id
        comp.pos_constr = self._pos_constr
        comp.named_constr = self._named_constr
        return comp


class ConfDefinitionBuilder(Builder):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(ConfDefinitionBuilder, self).__init__()
        self._id: CompId | None = None
        self._properties: DictDefinition | None = None

    def id_(self, id_: str) -> "ConfDefinitionBuilder":
        self._id = uuid.UUID(id_)
        return self

    def properties(self, properties: DictDefinition) -> "ConfDefinitionBuilder":
        self._properties = properties
        return self

    def build(self) -> ConfigurationDefinition:
        conf = ConfigurationDefinition(self._id)
        conf.properties = self._properties
        return conf


class RefDefinitionBuilder(Builder):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(RefDefinitionBuilder, self).__init__()
        self._name: str | None = None
        self._ref_comp_id: CompId | None = None

    def name(self, name: str) -> "RefDefinitionBuilder":
        self._name = name
        return self

    def ref_comp_id(self, id_: str) -> "RefDefinitionBuilder":
        self._ref_comp_id = uuid.UUID(id_)
        return self

    def build(self) -> ReferenceDefinition:
        return ReferenceDefinition(self._name, self._ref_comp_id)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import asyncio
from unittest import TestCase

from galaxy.app.app import GalaxyApplication,               \
                           GalaxyAsyncApplication


class TestLaunch(TestCase):

    #def test_launch_with_galaxy_async_kernel(self):
    #    app = GalaxyAsyncApplication("conf/c0ba5d81-97f2-445f-be51-cabcf7a825fd.yml")
    #    asyncio.run(app.load_conf("conf/57966f2b-f157-42a0-b85f-a66e086b90f2.yml"))
    #    asyncio.run(app.run())
    #    self.assertEqual(True, True)
    
    def test_launch_with_galaxy_kernel(self):
        app = GalaxyApplication("conf/763afe11-8d4c-4464-a937-a057c6cd7b4e.yml")
        app.load_conf("conf/2b789f3d-78ec-4ccd-bfb2-c9b7e7fe7729.yml")
        app.run()
        self.assertEqual(True, True)
    
    #def test_launch_with_jupyter_kernel(self):
    #    app = GalaxyApplication("conf/2511dada-de54-457c-b88a-d1fa16c9d4f3.yml")
    #    app.load_conf("conf/f221ee90-be67-416d-813b-a51d3f1a95c2.yml")
    #    app.run()
    #    self.assertEqual(True, True)


if __name__ == "__main__":
    unittest.main()



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                            \
                abstractmethod
from transitions.core import Machine,                           \
                             EventData
from typing import Any,                                         \
                   TYPE_CHECKING
import pwd
import os
import traceback

from galaxy.app import constant
from galaxy.app.argparser import ArgParser
from galaxy.app.ioc.ioc import IOCManager
from galaxy.utils.base import Component,                        \
                              TimestampedState
from galaxy.kernel.kernel import Kernel,                        \
                                 GalaxyAsyncKernel
from galaxy.service.service import ServiceManager,              \
                                   ServiceAsyncManager,         \
                                   LogService,                  \
                                   LogAsyncService
from galaxy.error.error import BaseError
from galaxy.proc.proc import ProcessManager,                    \
                             ProcessAsyncManager

if TYPE_CHECKING:
    from galaxy.app.ioc.visitor import Visitor


class Application(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        super(Application, self).__init__()
        self._machine: ApplicationStateMachine = ApplicationStateMachine(self)
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    def accept(self, visitor: "Visitor"):
        visitor.visit(self)

    def __repr__(self) -> str:
        return "<Application(id='{}')>".format(self.id)


class ApplicationState(TimestampedState):
    """
    classdocs
    """

    def __init__(self, name: str, app: Application) -> None:
        """
        Constructor
        """
        super(ApplicationState, self).__init__(name=name)
        self.app: Application = app


class ApplicationNewState(ApplicationState):
    """
    classdocs
    """

    def __init__(self, app: Application) -> None:
        """
        Constructor
        """
        super(ApplicationNewState, self).__init__(constant.STATE_NEW, app)


class ApplicationRunningState(ApplicationState):
    """
    classdocs
    """

    def __init__(self, app: Application) -> None:
        """
        Constructor
        """
        super(ApplicationRunningState, self).__init__(constant.STATE_RUNNING, app)

    def enter(self, event_data):
        # The logger is not already configured. It will be configured in the app._start()
        print("The application {} is booting".format(self.app))
        #self.app.log.logger.debug("The application {} is booting".format(self.app))
        self.app._start()
        self.app.log.logger.debug("The application {} is running".format(self.app))
        super(ApplicationRunningState, self).enter(event_data)


class ApplicationStoppedState(ApplicationState):
    """
    classdocs
    """

    def __init__(self, app: Application) -> None:
        """
        Constructor
        """
        super(ApplicationStoppedState, self).__init__(constant.STATE_STOPPED, app)

    def enter(self, event_data: EventData) -> None:
        self.app.log.logger.debug("The application {} is stopping".format(self.app))
        self.app._stop()
        self.app.log.logger.debug("The application {} is stopped".format(self.app))
        super(ApplicationStoppedState, self).enter(event_data)


class ApplicationShutdownState(ApplicationState):
    """
    classdocs
    """

    def __init__(self, app: Application) -> None:
        """
        Constructor
        """
        super(ApplicationShutdownState, self).__init__(constant.STATE_SHUTDOWN, app)


class ApplicationTimeoutState(ApplicationState):
    """
    classdocs
    """

    def __init__(self, app: Application) -> None:
        """
        Constructor
        """
        super(ApplicationTimeoutState, self).__init__(constant.STATE_TIMEOUT, app)


class ApplicationStateMachine(object):
    """
    classdocs
    """

    def __init__(self, app: Application) -> None:
        """
        Constructor
        """
        self._app: Application = app
        self._init_states()
        self._init_machine()

    def _init_states(self) -> None:
        self.states: dict[str, ApplicationState] = {
                                                    constant.STATE_NEW: ApplicationNewState(self._app),
                                                    constant.STATE_RUNNING: ApplicationRunningState(self._app),
                                                    constant.STATE_STOPPED: ApplicationStoppedState(self._app),
                                                    constant.STATE_SHUTDOWN: ApplicationShutdownState(self._app),
                                                    constant.STATE_TIMEOUT: ApplicationTimeoutState(self._app)
                                                   }
        self._transitions: list[dict[str, str]] = [{
                                                    "trigger": "run",
                                                    "source": constant.STATE_NEW,
                                                    "dest": constant.STATE_RUNNING
                                                   },
                                                   {
                                                    "trigger": "stop",
                                                    "source": constant.STATE_RUNNING,
                                                    "dest": constant.STATE_STOPPED
                                                   },
                                                   {
                                                    "trigger": "run",
                                                    "source": constant.STATE_STOPPED,
                                                    "dest": constant.STATE_RUNNING
                                                   },
                                                   {
                                                    "trigger": "shutdown",
                                                    "source": constant.STATE_STOPPED,
                                                    "dest": constant.STATE_SHUTDOWN
                                                   }]

    def _init_machine(self) -> None:
        self.machine: Machine = Machine(model=self._app,
                                        states=[state for state in self.states.values()],
                                        transitions=self._transitions,
                                        initial=self.states[constant.STATE_NEW])


class GalaxyApplication(Application):
    """
    classdocs
    """

    def __init__(self, comp_file: str) -> None:
        super(GalaxyApplication, self).__init__()
        self.ioc: IOCManager = IOCManager(self)
        self.kernel: Kernel | None = None
        self.service: ServiceManager | None = None
        self.proc: ProcessManager | None = None

        try:
            self.ioc.load_comp(comp_file)
        except BaseError as e:
            msg = "{} : {}".format(e, e.internal) if e.internal is not None else str(e)
            print(msg)
        except Exception as e:
            print(e)

    def load_conf(self, conf_file: str) -> None:
        try:
            self.ioc.load_conf(conf_file)
        except BaseError as e:
            msg = "{} : {}".format(e, e.internal) if e.internal is not None else str(e)
            print(msg)
        except Exception as e:
            print(e)

    def _start(self) -> None:
        self.service.load_all()
        self.proc.load_all()
        self.kernel.load()
        self.service.start()
        self.proc.start()
        self.kernel.run()

    def _stop(self) -> None:
        self.proc.stop()
        self.service.stop()
        self.kernel.stop()

    def __repr__(self) -> str:
        return "<GalaxyApplication(id='{}')>".format(self.id)


def start_app(*params: Any):
    username = pwd.getpwuid(os.getuid()).pw_name
    print("Launching program {} by {}".format(params[0], username))
    try:
        parser = ArgParser(params[0])
        args = parser.parse_args(params[1:])
        app = GalaxyApplication(args.comp)
        app.load_conf(args.conf)
        app.run()
    except Exception as e:
        print("An error occurred : {}".format(str(e)))
    print("End execution of the program {}".format(params[0]))


class GalaxyAsyncApplication(Application):
    """
    classdocs
    """

    def __init__(self, comp_file: str) -> None:
        super(GalaxyAsyncApplication, self).__init__()
        self.ioc = IOCManager(self)
        self.kernel: GalaxyAsyncKernel | None = None
        self.service: ServiceAsyncManager | None = None
        self.proc: ProcessAsyncManager | None = None

        try:
            self.ioc.load_comp(comp_file)
        except BaseError as e:
            msg = "{} : {}".format(e, e.internal) if e.internal is not None else str(e)
            print(msg)
        except Exception as e:
            print(traceback.format_exc())
            print(e)

    def load_conf(self, conf_file: str) -> None:
        self.ioc.load_conf(conf_file)

    def _start(self) -> None:
        self.kernel.load()
        self.kernel.run_async(self.service.load_all())
        self.kernel.run_async(self.service.start())
        self.kernel.run_async(self.proc.load_all())
        self.kernel.run_async(self.proc.start())
        self.kernel.run()

    def _stop(self) -> None:
        self.kernel.run_async(self.proc.stop())
        self.kernel.run_async(self.service.stop())
        self.kernel.stop()

    def __repr__(self) -> str:
        return "<GalaxyApplication(id='{}')>".format(self.id)


def start_async_app(*params: Any):
    username = pwd.getpwuid(os.getuid()).pw_name
    try:
        parser = ArgParser(params[0])
        args = parser.parse_args(params[1:])
        print("Launching program {} by {}".format(params[0], username))
        app = GalaxyAsyncApplication(args.comp)
        app.load_conf(args.conf)
        app.run()
    except Exception as e:
        print(traceback.format_exc())
        print("An error occurred : {}".format(str(e)))
    print("End execution of the program {}".format(params[0]))


class SimpleApplication(Application):
    """
    classdocs
    """

    def __init__(self, comp_file: str) -> None:
        super(SimpleApplication, self).__init__()
        self.ioc = IOCManager(self)
        self.service = None
        try:
            self.ioc.load_comp(comp_file)
        except BaseError as e:
            msg = "{} : {}".format(e, e.internal) if e.internal is not None else str(e)
            print(msg)
        except Exception as e:
            print(e)

    def load_conf(self, conf_file: str) -> None:
        try:
            self.ioc.load_conf(conf_file)
        except BaseError as e:
            msg = "{} : {}".format(e, e.internal) if e.internal is not None else str(e)
            print(msg)
        except Exception as e:
            print(e)

    def _start(self) -> None:
        self.service.load_all()

    def _stop(self) -> None:
        self.service.stop()

    def __repr__(self) -> str:
        return "<SimpleApplication(id='{}')>".format(self.id)



#  Copyright (c) 2023 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
from unittest import TestCase

from galaxy.app.app import SimpleApplication,           \
                           GalaxyAsyncApplication


class TestSimpleApp(TestCase):

    #def test_launch_simple_app(self) -> None:
    #    app = SimpleApplication("conf/a4102a60-440d-4a89-9668-23c12c4edf66.yml")
    #    app.load_conf("conf/8405a659-359b-4902-886f-ac6984c63cd0.yml")
    #    app.run()
    #    self.assertEqual(True, True)

    def test_launch_simple_galaxy_app(self) -> None:
        app = GalaxyAsyncApplication("conf/878826a5-431d-4565-8633-3e60977b0236.yml")
        app.load_conf("conf/47161190-b4af-4d8e-afd7-7d30550d4b66.yml")
        app.run()
        self.assertEqual(True, True)

if __name__ == "__main__":
    unittest.main()



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from os import path
import yaml
from abc import ABC,                                                    \
                abstractmethod
from typing import Any,                                                 \
                   TYPE_CHECKING
from collections import OrderedDict

from galaxy.app import constant
from galaxy.app.ioc.factory import DefinitionFactory
from galaxy.error.io import IOFileNotExistingError,                     \
                            IOFileWrongFormatOrMalformedError
from galaxy.utils.type import CompId

if TYPE_CHECKING:
    from galaxy.app.ioc.definition import ComponentDefinition


def convert(scope_str: str) -> str:
    if scope_str == "prototype":
        return constant.PROTOTYPE
    elif scope_str == "singleton":
        return constant.SINGLETON
    else:
        raise Exception("Can not handle scope {}".format(scope_str))


yaml_mappings: dict[str, str] = {
                                 "str": "types.StringType",
                                 "unicode": "types.UnicodeType",
                                 "int": "types.IntType",
                                 "long": "types.LongType",
                                 "float": "types.FloatType",
                                 "decimal": "decimal.Decimal",
                                 "bool": "types.BooleanType",
                                 "complex": "types.ComplexType",
                                 "list": "types.ListType",
                                 "tuple": "types.TupleType",
                                 "dict": "types.DictType",
                                }


def get_string(value: Any) -> Any:
    """
    classdocs
    """

    try:
        return str(value)
    except UnicodeEncodeError:
        return value


class CompositionParser(ABC):
    """
    classdocs
    """

    @abstractmethod
    def parse(self, comp_file: str) -> dict[CompId, "ComponentDefinition"]:
        raise NotImplementedError("Should implement parse()")


class ConfigurationParser(ABC):
    """
    classdocs
    """

    @abstractmethod
    def parse(self, comp_file: str) -> dict[CompId, "ConfigurationDefinition"]:
        raise NotImplementedError("Should implement parse()")


class YamlCompositionParser(CompositionParser):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(YamlCompositionParser, self).__init__()

    def parse(self, comp_file: str) -> dict[CompId, "ComponentDefinition"]:
        if not path.exists(comp_file):
            raise IOFileNotExistingError(comp_file)

        factory = DefinitionFactory()
        try:
            with open(comp_file) as f:
                doc = yaml.load(f, Loader=yaml.FullLoader)
        except Exception as e:
            raise IOFileWrongFormatOrMalformedError(comp_file, None, e)
        comps = factory.create_components(doc)

        #visitor = PrintVisitor()
        #for comp in comps:
        #    comp.accept(visitor)

        return OrderedDict({comp.id: comp for comp in comps})


class YamlConfigurationParser(ConfigurationParser):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(YamlConfigurationParser, self).__init__()

    def parse(self, conf_file: str) -> dict[CompId, "ConfigurationDefinition"]:
        if not path.exists(conf_file):
            raise IOFileNotExistingError(conf_file)

        factory = DefinitionFactory()
        try:
            with open(conf_file) as f:
                doc = yaml.load(f, Loader=yaml.FullLoader)
        except Exception as e:
            raise IOFileWrongFormatOrMalformedError(conf_file, None, e)
        confs = factory.create_configurations(doc)

        #visitor = PrintVisitor()
        #for conf in confs:
        #    conf.accept(visitor)

        return OrderedDict({conf.id: conf for conf in confs})



#  Copyright (c) 2023 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import logging
from argparse import ArgumentParser
from argparse import RawTextHelpFormatter


class ArgParser(object):
    """
    classdocs
    """

    def __init__(self, exec_name) -> None:
        """
        Constructor
        """
        self.exec_name = os.path.basename(exec_name)
        self.short_desc = """
The application Galaxy
"""
        self.app_license = """{}
Copyright (c) 2023 bastien.saltel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

USAGE :
""".format(self.short_desc)
        self._parser = ArgumentParser(prog=self.exec_name, description=self.app_license, formatter_class=RawTextHelpFormatter)
        self._parser.add_argument("-c", "--comp", dest="comp", help="set the composition file")
        self._parser.add_argument("-f", "--conf", dest="conf", help="set the configuration file")
        self._parser.add_argument("-V", "--version", action="version", version="{prog}s {version}".format(prog="%(prog)", version="1.0"), help="show version")

    def parse_args(self, params):
        try:
            args = self._parser.parse_args(params)
        except Exception as e:
            print("An error occurred while parsing the arguments : {}".format(str(e)))
            raise
        if not os.path.isfile(args.comp):
            print("The composition file '{}' does not exist".format(args.comp))
            self._parser.print_help()
            raise
        if not os.path.isfile(args.conf):
            print("The configuration file '{}' does not exist".format(args.comp))
            self._parser.print_help()
            raise
        return args
